(* 6) AST-to-byte code compiler *)

open Ast
open Bytecode

module StringMap = Map.Make(String)

(* Symbol table: Information about all the names in scope *)
type env = {
	function_index : int StringMap.t; (* Index for each function *)
	global_index : int StringMap.t; (* "Address" for global vars *)
	local_index : int StringMap.t; (* FP offset for args, locals *)
}

(* enum: get a list of things, and 2 ints (step, start), *)
(* and produce another list of tuples (int, thing) *)
(* if start is 1, and step is 1, the list a,b,c *)
(* will become (a,1),(b,2)(c,3) *)
(* val enum : int -> int -> 'a list -> (int * 'a) list = <fun> *)
(* aka. world's simplest memory allocator*)
 
(* this would have to change if we had different types of variables in the language, *)
(* but here all vars are ints, so step is passed as constant *)
(* and we assume all vars are of size step*)
let rec enum step start = function
		[] -> []
	| hd:: tl -> (start, hd) :: enum step (start + step) tl

(* string_map_pairs:StringMap 'a -> (int * 'a) list -> StringMap 'a *)
(* val string_map_pairs :
  'a StringMap.t -> 
		('a * StringMap.key) 
	list -> 'a StringMap.t = <fun>
# *)
let string_map_pairs map pairs =
	List.fold_left (fun m (i, n) -> StringMap.add n i m) map pairs

(** Translate a program in AST form into a bytecode program.
 Throw an exception if something is wrong, 
e.g., a reference to an unknown variable or function *)
let translate (globals, functions) =
	
	(* Allocate "addresses" / indexes  for each global variable *)
	let global_indexes =
		string_map_pairs StringMap.empty (enum 1 0 globals) 
	(* step = 1, start = 0 *)
	
	in 
	
	(* Assign "addresses" / indexes to function names; *)
	(* builtin functions like "print", "penup", "pendown", etc*)
	(* are special *)
	let built_in_functions =
		StringMap.add "print" (-1) StringMap.empty in
	let built_in_functions = 
		StringMap.add "pendown" (-2) built_in_functions  in
	let built_in_functions = 
		StringMap.add "penup" (-3) built_in_functions  in
	let built_in_functions = 
		StringMap.add "fwd" (-4) built_in_functions  in
	let built_in_functions = 
		StringMap.add "left" (-5) built_in_functions  in		
		
	let function_indexes = string_map_pairs built_in_functions
			(enum 1 1 (List.map (fun f -> f.fname) functions)) 
			
	in 
	
	(* Translate an AST function to a list of bytecode statements *)
	let translate env fdecl =
		
		(* Bookkeeping: FP offsets for locals and arguments *)
		let num_formals = List.length fdecl.formals
		and num_locals = List.length fdecl.locals
		
		
		and local_offsets = enum 1 1 fdecl.locals (* locals go down the stack, starting with at 1, 
		with stride 1, so local1 at 1, local2 at 2, and so on *)
		
		(* the args of the function do up the stack, starting at -2, with stride -1, *)
		(* so arg1 is at -2, arg2 is at -3, arg3 is at -4, and so on *)
		and formal_offsets = enum (-1) (-2) fdecl.formals in
		
		
		(* NOTE: use of 'with' as part of the env type *) 
		let env = { env with local_index = string_map_pairs
					StringMap.empty (local_offsets @ formal_offsets) } in
					
					
		let rec expr = function
				Literal i -> print_endline("Literal"); [Lit i]
			| Id s -> print_endline("Id");
				(* first, try to find the identifier *)
				(* in the local vars or the function args *)
					(try [Lfp (StringMap.find s env.local_index)]
					
					(* if it is not among local vars and args, *)
					(* try global vars and other functions *)
					with Not_found -> try
								[Lod (StringMap.find s env.global_index)]
							(* not found among args, locals, globals, other funcs,*)
							(*  so not found at all anywhere, error *)
							with Not_found ->
									raise (Failure ("undeclared variable " ^ s)))
			
			(* evaluate e1, then e1, combine them *)
			| Binop (e1, op, e2) -> print_endline("Binop");expr e1 @ expr e2 @ [Bin op]
			
			| Assign (s, e) -> print_endline("Assign");expr e @
					
					(* try to assign to local or arg first *)
					(try [Sfp (StringMap.find s env.local_index)]
					with Not_found -> try
								(* try to assign to global if local or arg was not found*)
								(* under the given name *)
								(* TEST: assign value to a function in the global index??? *)
								[Str (StringMap.find s env.global_index)]
							with Not_found ->
									raise (Failure ("undeclared variable " ^ s)))
			
			(* call a function by its name, supply list of inputs *)
			(* NOTE: inputs evaluated right to left!!! *)
			(* beware of side-effects *)
			| Call (fname, actuals) -> print_endline((string_of_int num_formals)^"Call"^(string_of_int (List.length actuals)));
					
					(try
						let idx = StringMap.find fname env.function_index in
						(* evaluate each of the args, form a list of the *)
						(* evaluated args and append a jump to the function location *)
						(* as the last instruction in the list *)
						(List.concat (List.map expr (List.rev actuals))) @
						[Jsr idx ] @
						(* NOTE: special addition for built-in functions *)
						(* before I added this, print(); followed by print(); would loop for ever *)
						(if (idx < 0 && ((List.length actuals) = 0)) then 
							[Lit 0;] else [])
					with Not_found ->
							raise (Failure ("undefined function " ^ fname)))
			| Noexpr -> print_endline("Noexpr");[] (* so that the for loop works *)
		
		(* Translate a statement *)
		in let rec stmt = function
				Block sl -> List.concat (List.map stmt sl)
				(* if it is block consisting of list of statements, for each item *)
				(* in the list, process it and the append to the end *)
				
			| Expr e -> print_endline("expr+drop");expr e @ [Drp] (* Discard result *)
			
			(* NOTE: here we use NUM_FORMALS (the number of args passed)*)
			(* to calculate where to put the end result of *)
			(* evaluting the expr, because this location is where*)
			(* the caller will look for the result *)
			| Return e -> expr e @ [Rts num_formals]
			
			| If (p, t, f) -> 
				  (* generate the byte code instructions for the true case *)
					let t' = stmt t 
					
					(* generate the byte code instructions for the false case *)
					and f' = stmt f in
					
					(* generate the byte code instructions *)
					(* for the condition check, *)
					(* put it first in the list *)
					(* of generated byte code *)
					(* instructions to be returned *)
					expr p 
					
					@ 
					
					(* jump over the then/true case code if needed *)
					[Beq(2 + List.length t')] 
					
					@
					
					(* the byte code for the true/then case *) 
					t' 
					
					@
					
					(* branch unconditionally past the false case *)
					(* cause we can reach this place only if we already executed *)
					(* the true case byte code, we don't want to execute both, right? *)	  
					[Bra(1 + List.length f')] 
					
					@ 
					
					(* last, append the byte codes for the false case*)
					(* to the list of byte code instructions to be returned *)
					f'
			
			| For (e1init, e2stop, e3change, blk) -> 
				  (* Rewrite into a while statement *)
					stmt (
						Block(
						[
							Expr(e1init); 
						  While(e2stop, Block([blk; Expr(e3change)]))
						]
						))
		
			| While (e, b) ->
					
					(* generate byte codes for the block of code *)
					(* inside the while loop, *)
					(* we need to know its length, that is*)
					(* how many byte code instructions *)
					(* it translated into, so that we can jump over them*)
					(* using the correct offset in the *)
					(* jump relative bytecode instruction *)
					let b' = stmt b 
					
					(* generate the byte code to evaluate the predicate *)
					and e' = expr e 
					
					in
		
					(* build the list of byte codes to be returned for while loop *)
					 			
					(* standard way of doing things: branch unconditionally *)
					(* past the code, avoids 2 branch instructions *)
					[Bra (1 + List.length b')] 
					
					@ 
					
					(* the byte code instuctions for the code inside the while loop*)
					(* come next *)
					b'
					 
					@ 
					
					(* the byte code for evaluating the loop predicate comes after*)
					(* the block of whatever was inside the while *)
					(* so basically we are turning a while-do *)
					(* into an equivalent do-while because *)
					(* do-while is more efficient *)
					e' 
					
					@
					
					(* after evaluating the loop predicate *)
					(* if it is true, we need to go back to the beginning*)
					(* of the block of code and run it *)
					(* since we generated the byte codes for the block and the*)
					(* for evaluating the predicate beforehand, *)
					(* we know exactly where to jump *)
					(* so that we land right in the beginning *)
					(* of the byte codes for the while-loop block *)
					[Bne (- (List.length b' + List.length e'))]
		
		(* Translate a whole function *)
		in 
		
		(* Entry: allocate space for locals *)
		[Ent num_locals] 
		
		@ 
		
		(* generate the byte code for the function body *)
		stmt (Block fdecl.body) 
		
		@ 
		(* put the result of executing the function *)
		(* where the caller expects it *)
		[Lit 0; Rts num_formals] (* Default = return 0 *)
	
	in 
	
	(* now that we have generated the byte code for the functions, *)
	(* we know how many instruction everything is, *)
	(* so we can calculate the indexes / offsets / addresses*)
	(* and create the env as it is*)
	(* when we enter the main function *)
	(* in the very beginning *)
	let env = {
		function_index = function_indexes;
		global_index = global_indexes;
		local_index = StringMap.empty } 
		
	in
	
	(* Code executed to start the program: Jsr main; halt *)
	let entry_function = try
			(* very classic, x86 does just that: *)
			(* in the beginnin, just to main, then halt *)
			(* well, x86 does some extra stuff like setting *)
			(* the exit code and such, which we omit to do here *)
			[Jsr 
			(* lookup the address / index / location of the main function *)
			(* in our bytecode array *)
			(StringMap.find "main" function_indexes); 
			(* classic: call main, then exit *)
			Hlt]
	
		(* TEST: make sure we can;t have 2 function with the same name *)
		(* TEST: try 2 function with same name, diff arg list *)
		(* TEST: try a global var and a function having the same name *)
		(* TEST: try to declare a function inside a function *)		
		with Not_found -> raise (Failure ("no \"main\" function"))
	
	in
	
	(* Compile the functions *)
	(* func_bodies is a list of lists of bytecode statements *)
	let func_bodies = 
		(* put the bytecode for startig the program *)
		(* which means 1 jump instr to the offset where*)
		(* main is in our bytecode array and then exit the program *)
		entry_function (* this is a list of 2 byte code instr: jump and halt *)
		
		:: (* this operator is making the entry function list the first *)
		(* element in the list of lists of bytecode statements *)
		(* each list in this list of lists containts the bytecode instuctions*)
		(* for a given function *)
		
		(* put the bytecode for the other, user-defined,*)
		(* non-mandatory like main, functions next *)
		List.map (translate env) functions 
		
	in
	
	(* Calculate function entry points by adding their lengths *)
	(* fun_offset_list is a list of ints which represent offsets *)
	let (fun_offset_list, _) = List.fold_left
			(fun (l, i) f -> (i :: l, (i + List.length f))) ([],0)
			 	func_bodies 
	in
	
	(* replacing labels/indexes with actual offsets *)
	(* after we have generated all bytecode and we know *)
	(* what these offsets are *)
	
	let func_offset =
		 Array.of_list 
		(List.rev fun_offset_list) (* NOTE: we reverse the list *)
		
	in
	(* Concatenate the compiled functions *)
		(* and replace the function indexes in  *)
		(* Jsr statements with PC values *)   
		
	{ num_globals = List.length globals;
		                                       
		text = Array.of_list ( List.map (function Jsr i 
			(* NOTE: Jsr with negatives are the special built in function*)
			(* leave them alone *)
			(* only calculate Program Counter (~ line number) *)
			(* values for function coming from user-defined code *)
			when i > 0 -> Jsr func_offset.(i)
			| _ as s -> s) (* no change for non-Jsr byte codes *)
			
			(* apply this to all generated byte instruction *)
			(* for all user defined function *)
			 (List.concat func_bodies)
			
			)
	}
